# refer https://duckdb.org/docs/sql/duckdb_table_functions.html
has_index_query = "SELECT indexname FROM pg_indexes where schemaname = ? and tablename = ? and indexname = ?"  # noqa: E501,B950
has_table_query = "select table_name from information_schema.tables where table_schema = ? and table_name = ?"  # noqa: E501,B950
has_sequence_query = "select sequencename from pg_sequences where schemaname = ? and sequencename = ?"  # noqa: E501,B950
get_tables = (
    "select tablename from pg_tables where schemaname = ?"  # noqa: E501,B950
)
get_schemas = "SELECT schema_name FROM information_schema.schemata where catalog_name not in ('system','temp')"  # noqa: E501,B950
get_views = "select view_name from duckdb_views where schema_name = ?"  # noqa: E501,B950
get_view_sql = "select sql from duckdb_views where schema_name = ? and view_name= ?"  # noqa: E501,B950
get_constraints = "select constraint_text as name,constraint_column_names as column_names from duckdb_constraints where schema_name = ? and table_name = ? and constraint_type = ?"  # noqa: E501,B950
get_temp_views = "select view_name from duckdb_views where temporary = true;"  # noqa: E501,B950
get_temp_tables = "select table_name from information_schema.tables where table_schema = ? and table_type like '%TEMPORARY%';"  # noqa: E501,B950
get_sequences = "select sequencename from pg_sequences where schemaname = ?"  # noqa: E501,B950
get_indexes = "select index_name, sql, is_unique from duckdb_indexes where schema_name = ? and table_name = ?"  # noqa: E501,B950
get_columns = "SET schema '{}'; PRAGMA table_info('{}')"
get_check_constraint = "select constraint_text as name,expression as sqltext from duckdb_constraints where schema_name = ? and table_name = ? and constraint_type = 'CHECK'"  # noqa: E501,B950



import pytest

from radio_duck import OperationalError, ProgrammingError, connect, db_types
from radio_duck.connection_test import http_server_port


def test_no_query():
    from http_server_mock import HttpServerMock

    app = HttpServerMock(__name__)

    @app.route("/v1/sql/", methods=["POST"])
    def index():
        return "{}"

    with app.run("localhost", http_server_port):
        try:
            with connect(
                host="localhost",
                port=http_server_port,
                api="/v1/sql/",
                scheme="http",  # noqa: E501
            ) as conn:
                with conn.cursor() as cursor:
                    cursor.execute("   ")
                    pytest.fail(
                        "Gave empty query but did not error out. failing test"
                    )  # noqa: E501
        except ProgrammingError as e:
            assert "query is empty" in e.msg.lower()


def test_cursor_description():
    from http_server_mock import HttpServerMock

    app = HttpServerMock(__name__)

    @app.route("/v1/sql/", methods=["POST"])
    def index():
        return (
            """{
              "schema": [
                "STRING", "NUMBER"
              ],
              "columns": [
                "duck_type", "total"
              ],
              "rows": [["mallard", 1], ["marbled_duck", 2]]
            }
        """,
            200,
        )

    with app.run("localhost", http_server_port):
        conn = connect(
            host="localhost",
            port=http_server_port,
            api="/v1/sql/",
            scheme="http",
        )
        cursor = conn.cursor()
        cursor.execute("select duck_type from pond")
        cursor_desc = cursor.description

        assert 2 == len(
            cursor_desc
        ), "expecting 2 descrirption for 2 column selected"  # noqa: E501
        for string_col_desc in cursor_desc:
            assert 7 == len(
                string_col_desc
            ), "expecting sequence of size 7 for each description"
        string_col_desc = cursor_desc[0]
        assert (
            string_col_desc[0] == "duck_type"
        ), "expected col name in cursor description to be 'duck_type'"
        assert string_col_desc[1] == db_types.get_type_code(
            "STRING"
        ), "expected col type in cursor description to be 'STRING'"

        num_col_desc = cursor_desc[1]
        assert (
            num_col_desc[0] == "total"
        ), "expected col name in cursor description to be 'total'"
        assert num_col_desc[1] == db_types.get_type_code(
            "NUMBER"
        ), "expected col type in cursor description to be 'NUMBER'"


def test_cursor_iteration():
    from http_server_mock import HttpServerMock

    app = HttpServerMock(__name__)

    @app.route("/v1/sql/", methods=["POST"])
    def index():
        return (
            """{
              "schema": [
                "STRING"
              ],
              "columns": [
                "some_number"
              ],
              "rows": [["one"], ["two"], ["three"],
              ["four"], ["five"], ["six"]]
            }
        """,
            200,
        )

    with app.run("localhost", http_server_port):
        with connect(
            host="localhost",
            port=http_server_port,
            api="/v1/sql/",
            scheme="http",
        ) as conn:
            with conn.cursor() as cursor:
                assert (
                    -1 == cursor.rowcount
                ), "testing rowcount before execution"  # noqa: E501
                cursor.execute("select some_number from numbers")
                assert 6 == cursor.rowcount, "testing rowcount after execution"
                # set & test fetch size
                cursor.arraysize = 2
                assert 2 == cursor.arraysize

                row = cursor.fetchone()
                assert row is not None
                assert "one" == row[0]

                twoRows = cursor.fetchmany()
                assert twoRows is not None
                assert cursor.arraysize == len(twoRows)
                assert "two" == twoRows[0][0]
                assert "three" == twoRows[1][0]

                remaining = cursor.fetchall()
                assert remaining is not None
                assert len(remaining) == 3
                assert "four" == remaining[0][0]
                assert "five" == remaining[1][0]
                assert "six" == remaining[2][0]

                # no more left
                assert cursor.fetchone() is None
                anymore = [cursor.fetchall(), cursor.fetchmany(2)]
                for am in anymore:
                    assert am is not None
                    assert 0 == len(am)


def test_cursor_close():
    from http_server_mock import HttpServerMock

    app = HttpServerMock(__name__)

    @app.route("/v1/sql/", methods=["POST"])
    def index():
        return (
            """{
              "schema": [
                "STRING"
              ],
              "columns": [
                "some_number"
              ],
              "rows": [["one"]]
            }
        """,
            200,
        )

    with app.run("localhost", http_server_port):
        with connect(
            host="localhost",
            port=http_server_port,
            api="/v1/sql/",
            scheme="http",
        ) as conn:
            cursor = conn.cursor()
            try:
                cursor.execute("select some_number from numbers")
            finally:
                cursor.close()
                try:
                    cursor.execute("select total from bank")
                except Exception as e:
                    assert "already closed" in e.msg
                try:
                    cursor.fetchone()
                except Exception as e:
                    assert "already closed" in e.msg
                try:
                    cursor.fetchall()
                except Exception as e:
                    assert "already closed" in e.msg
                try:
                    cursor.fetchmany(2)
                except Exception as e:
                    assert "already closed" in e.msg


def test_for_non200response():
    from http_server_mock import HttpServerMock

    app = HttpServerMock(__name__)

    @app.route("/v1/sql/", methods=["POST"])
    def index():
        raise Exception("internal error")

    @app.route("/v2/sql/", methods=["POST"])
    def index1():
        return "{}", 408

    with app.run("localhost", http_server_port):
        with connect(
            host="localhost",
            port=http_server_port,
            api="/v1/sql/",
            scheme="http",
        ) as conn:
            with conn.cursor() as cursor:
                try:
                    cursor.execute("select * from pond")
                    pytest.fail(
                        "api endpoint throws 500 but cursor.execute() did not error out . failing test"  # noqa: E501,B950
                    )
                except Exception as e:
                    assert isinstance(e, OperationalError)
                    assert (
                        "failed to execute query. response status: 500"
                        in e.msg
                    )  # noqa: E501
        with connect(
            host="localhost",
            port=http_server_port,
            api="/v2/sql/",
            scheme="http",
        ) as conn:
            with conn.cursor() as cursor:
                try:
                    cursor.execute("select * from")
                    pytest.fail(
                        "api endpoint throws 400, but cursor.execute() did not error out . failing test"  # noqa: E501,B950
                    )
                except Exception as e:
                    assert isinstance(e, ProgrammingError)
                    assert (
                        "failed to execute query. response status: 408"
                        in e.msg
                    )  # noqa: E501


def test_for_bad_response_json():
    from http_server_mock import HttpServerMock

    app = HttpServerMock(__name__)

    @app.route("/v1/sql/", methods=["POST"])
    def index():
        return "this is not a json!", 200

    with app.run("localhost", http_server_port):
        with connect(
            host="localhost",
            port=http_server_port,
            api="/v1/sql/",
            scheme="http",
        ) as conn:
            with conn.cursor() as cursor:
                try:
                    cursor.execute("select * from pond")
                    pytest.fail(
                        "api returned bad json,"
                        " but cursor.execute() did not error out. failing test"
                    )
                except Exception as e:
                    assert isinstance(e, OperationalError)
                    assert "could not deserialize response" in e.msg


def test_fetch_before_execute():
    from http_server_mock import HttpServerMock

    app = HttpServerMock(__name__)

    @app.route("/v1/sql/", methods=["POST"])
    def index():
        return "{}", 200

    with app.run("localhost", http_server_port):
        with connect(
            host="localhost",
            port=http_server_port,
            api="/v1/sql/",
            scheme="http",
        ) as conn:
            with conn.cursor() as cursor:
                try:
                    cursor.fetchone()
                except Exception as e:
                    assert isinstance(e, ProgrammingError)
                try:
                    cursor.fetchall()
                except Exception as e:
                    assert isinstance(e, ProgrammingError)
                try:
                    cursor.fetchmany()
                except Exception as e:
                    assert isinstance(e, ProgrammingError)


@pytest.mark.integration_test
def test_with_server():
    # start radio-duck server for this test
    # https://github.com/jaihind213/radio-duck
    with connect(
        host="localhost", port=8000, api="/v1/sql/", scheme="http"
    ) as conn:
        with conn.cursor() as cursor:
            # good query1
            cursor.execute(
                "select total from pond where duck_type = ? ", ["mighty_duck"]
            )
            assert cursor.rowcount > 0, "testing rowcount after execution"
            row = cursor.fetchone()
            assert row is not None
            assert row[0] > 1

            # reuse cursor
            try:
                # bad query1 where ? is missing but parameters given
                cursor.execute(
                    "select total from pond where duck_type", ["mighty_duck"]
                )
            except Exception as e:
                assert "Invalid Input Error" in e.msg
                assert isinstance(e, ProgrammingError)
            try:
                # bad query2
                cursor.execute("select column from ")
            except Exception as e:
                assert "Parser Error" in e.msg
                assert isinstance(e, ProgrammingError)

            # good query again after bad queries
            cursor.execute(
                "select total from pond where duck_type = ? ", ["a_whale"]
            )  # noqa: E501
            assert cursor.rowcount == 0, "testing rowcount after execution"
            row = cursor.fetchone()
            assert row is None



import time

import pytest

from radio_duck import InterfaceError, OperationalError, connect

http_server_port = 9001


def test_bad_scheme():
    try:
        connect(
            host="localhost",
            port=http_server_port,
            api="/v1/sql/",
            scheme="grpc",
        )
        pytest.fail(
            "Gave wrong scheme but connection created successfully & did not throw error. failing test"  # noqa: E501,B950
        )
    except InterfaceError as e:
        assert "driver only supports http scheme" in e.msg.lower()


def test_bad_url():
    try:
        connect(
            host="mandolorian_rocks",
            port=http_server_port,
            api="/v1/sql/",
            scheme="http",
        )
        pytest.fail(
            "Gave wrong host,port but connection created successfully ! failing test"  # noqa: E501,B950
        )
    except Exception as expected:  # noqa: F841
        pass


def test_connection_open_close():
    from http_server_mock import HttpServerMock

    app = HttpServerMock(__name__)

    @app.route("/v1/sql/", methods=["POST"])
    def index():
        return "{}"

    with app.run("localhost", http_server_port):
        conn = connect(
            host="localhost",
            port=http_server_port,
            api="/bad_enpoint",
            scheme="http",
        )
        with conn.cursor() as cursor:
            try:
                cursor.execute("select * from pond")
                pytest.fail(
                    "Gave wrong endpoint, expected error. failing test"
                )  # noqa: E501
            except Exception as e:
                # if we get 404,
                # => connection was opened successfully.
                # 404 is valid response.
                assert "404" in e.msg
        conn.close()
        try:
            conn.cursor()
            pytest.fail("Created cursor on closed connection. failing test")
        except Exception as expected:  # noqa: F841
            pass


def test_connection_timeout():
    from http_server_mock import HttpServerMock

    app = HttpServerMock(__name__)

    timeout_sec = 3

    @app.route("/v1/sql/", methods=["POST"])
    def index():
        time.sleep(timeout_sec + 2)
        return "{}"

    with app.run("localhost", http_server_port):
        with connect(
            host="localhost",
            port=http_server_port,
            api="/v1/sql/",
            scheme="http",
            timeout_sec=timeout_sec,
        ) as conn:
            with conn.cursor() as cursor:
                try:
                    cursor.execute("select * from pond")
                    pytest.fail("should timeout but did not. failing test")
                except OperationalError as e:
                    assert "timed out" in str(e)



# https://peps.python.org/pep-0249/#exceptions


class Error(Exception):
    def __init__(self, *args, **kwargs):
        self.msg = kwargs.get("msg", "")
        self.response_status = kwargs.get("response_status", -1)

    def __str__(self):
        details = "" if self.__cause__ is None else repr(self.__cause__)
        return (
            f"radio_duck_error: msg {self.msg}, "
            f"response_status: {self.response_status}. "
            f"details: {details}"
        )


class Warning(Exception):
    def __init__(self, *args, **kwargs):
        self.msg = kwargs.get("msg", "")


class InterfaceError(Error):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class DatabaseError(Error):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class InternalError(DatabaseError):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class OperationalError(DatabaseError):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class ProgrammingError(DatabaseError):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class IntegrityError(DatabaseError):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class DataError(DatabaseError):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class NotSupportedError(DatabaseError):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


__all__ = [
    "Error",
    "DataError",
    "ProgrammingError",
    "IntegrityError",
    "InterfaceError",
    "OperationalError",
    "InternalError",
    "DatabaseError",
    "NotSupportedError",
]



import http.client
import json
import logging
from functools import wraps
from typing import Any, List, Optional, Union

from radio_duck.db_types import get_type_code
from radio_duck.exceptions import (
    InterfaceError,
    NotSupportedError,
    OperationalError,
    ProgrammingError,
)

connect_close_resource_msg = "connect_resource_closure"


def check_closed(f):
    """
    Decorator that checks if connection/cursor is closed.
    """

    @wraps(f)
    def g(self, *args, **kwargs):
        if self.closed:
            raise InterfaceError(
                msg=f"[{connect_close_resource_msg}]: {self.__class__.__name__} already closed"  # noqa: E501,B950
            )
        return f(self, *args, **kwargs)

    return g


class Connection(object):
    """
    A DB-API 2.0 (PEP 249) connection.

    Do not create this object directly; use globals.connect().
    """

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.close()

    def __init__(self, *args, **kwargs):
        self.host = kwargs.get("host", "specify_host")
        self.port = kwargs.get("port", 8000)
        self.scheme = kwargs.get("scheme", "http")
        self.timeout_sec = kwargs.get("timeout_sec", 10)  # todo
        self.closed = False
        self.api = kwargs.get("api", "/v1/sql/")
        if self.scheme == "http":
            self._http_connection = http.client.HTTPConnection(
                self.host, self.port, timeout=self.timeout_sec
            )
            try:
                logging.info(
                    "connecting to radio_duck... {}://{}:{}{}".format(
                        self.scheme, self.host, self.port, self.api
                    )
                )
                self._http_connection.connect()
                logging.info("connected to radio_duck")
            except Exception as e:
                raise OperationalError(
                    f"unable to connect to database: {e}"
                ) from e  # noqa: E501
        else:
            raise InterfaceError(
                msg="driver only supports http scheme for now"
            )  # noqa: E501

    def close(self):
        logging.debug("closing connection to radio_duck")
        self._http_connection.close()
        self.closed = True
        logging.info("closed connection to radio_duck")

    @check_closed
    def commit(self):
        # for transaction. do everything in a single cursor operation.
        # limited support at this time
        raise NotSupportedError(
            msg="do everything in a single cursor execute. 'begin;......;commit();'"  # noqa: E501
        )

    def rollback(self):
        # for transaction. do everything in a single cursor operation.
        # limited support at this time
        raise NotSupportedError(
            msg="do everything in a single cursor execute. 'begin;......;commit();'"  # noqa: E501
        )

    @check_closed
    def cursor(self, *args, **kwargs):
        return Cursor(self, **kwargs)

    @property
    def http_connection(self):
        return self._http_connection


class Cursor(object):
    """
    A DB-API 2.0 (PEP 249) cursor.

    Do not create this object directly; use Connection.cursor().
    """

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.close()

    def __init__(self, connection: Connection):
        self.closed = False
        self._result = None
        self._connection = connection
        self._rowcount = -1
        self._arraysize = 1
        self._index = -1
        logging.debug("opened cursor to radio_duck")

    @property
    def connection(self) -> Connection:
        """
        Get the connection associated with this cursor.

        This is an optional DB-API extension.
        """
        return self._connection

    @property
    def rowcount(self):
        if self._result is None:
            return -1
        return len(self._result["rows"])

    def callproc(self, procname, *args):
        raise NotSupportedError(msg="callproc not supported on cursor")

    def close(self):
        # free up the resources
        self._result = None
        self._rowcount = -1
        self._index = -1
        self.closed = True
        logging.debug("closed cursor to radio_duck")

    @check_closed
    def execute(self, query: Union[bytes, str], parameters=None):
        """
        Execute a query.

        Parameters
        ----------
        query : bytes or str
            The query to execute.  Pass SQL queries as strings,
            (serialized) Substrait plans as bytes.
        parameters
            Parameters to bind.  Can be a Python sequence (to provide
            a single set of parameters).
        :raise OperationalError if unable to execute query
        :raise ProgrammingError if query is empty or invalid or improper(ex: table not found)  # noqa: E501,B950
        """
        if query is None or "" == query.strip():
            raise ProgrammingError(msg="query is empty")

        request = {
            "sql": query,
            "timeout": self._connection.timeout_sec,
            "parameters": parameters,
        }
        headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
        }

        http_response = None
        response_status = -1
        response_payload = None

        try:
            request_payload = json.dumps(request)
            self._connection.http_connection.request(
                "POST",
                self._connection.api,
                body=request_payload,
                headers=headers,
            )
            http_response = self._connection.http_connection.getresponse()
            response_status = http_response.status
            response_payload = http_response.read()
        except Exception as e:
            logging.error("error in querying server {}".format(e))
            raise OperationalError(
                msg=f"failed to execute query. response status {response_status}. response: {response_payload}"  # noqa: E501,B950
            ) from e
        finally:
            if http_response is not None:
                http_response.close()

        if response_status != 200:
            msg = (
                "failed to execute query. response status: "
                f"{response_status}. response_payload: {response_payload}"
            )
            if _is4xx(response_status):
                raise ProgrammingError(
                    msg=msg, response_status=response_status
                )
            else:
                raise OperationalError(
                    msg=msg, response_status=response_status
                )

        try:
            self._result = json.loads(response_payload.decode("utf-8"))
            self._index = 0
        except Exception as e:
            logging.error(
                "error in deserializing json response server {}".format(e)
            )
            raise OperationalError(  # noqa: E501,B950
                msg=f"Failed to execute query. could not deserialize response: {e}."  # noqa: E501,B950
            ) from e

    def executemany(self, query: Union[bytes, str], seq_of_parameters) -> None:
        raise NotSupportedError(
            msg="`executemany` is not supported, use `execute` instead"
        )

    @check_closed
    def fetchone(self) -> Optional[tuple]:
        if self._result is None:
            raise ProgrammingError(msg="cannot fetchone() before execute()")

        rows = self._result["rows"]
        if len(rows) == 0 or self._index < 0 or self._index >= len(rows):
            return None
        to_return = rows[self._index]
        self._index = self._index + 1
        return to_return

    @check_closed
    def fetchmany(self, size: Optional[int] = None) -> List[tuple]:
        if self._result is None:
            raise ProgrammingError(msg="cannot fetchmany before execute()")
        if size is None or size <= 0:
            size = self._arraysize
        # _result['rows'] is [], slice it to get elems
        next_elements = self._result["rows"][
            self._index : self._index + size  # noqa: E203,E501
        ]
        self._index = self._index + size
        return next_elements

    @check_closed
    def fetchall(self) -> List[tuple]:
        if self._result is None:
            raise ProgrammingError(msg="cannot fetchall before execute()")
        remaining = self._result["rows"][self._index :]  # noqa: E203
        self._index = len(self._result["rows"])
        return remaining

    def nextset(self):
        """Move to the next available result set (not supported)."""
        raise NotSupportedError(msg="cursor.nextset")

    @property
    def arraysize(self):
        return self._arraysize

    @arraysize.setter
    def arraysize(self, size):
        self._arraysize = size

    def setinputsizes(self, sizes):
        """Preallocate memory for the parameters (no-op)."""
        pass

    def setoutputsize(self, size, column=None):
        """Preallocate memory for the result set (no-op)."""
        pass

    @property
    def description(self):
        """
        This read-only attribute is a sequence of 7-item sequences.
        """
        if self.closed or self._result is None:
            return None

        columns_types = self._result.get("schema", [])
        column_names = self._result.get("columns", [])

        return self._get_description(columns_types, column_names)

    def _get_description(
        self, columns_types: List[str], column_names: List[str]
    ) -> Any:
        return [
            (
                column_name,
                get_type_code(column_type),
                None,
                None,
                None,
                None,
                True,
            )
            for column_name, column_type, in zip(  # noqa: B905,E501
                column_names, columns_types
            )
            # strict = True for zip not working
            # https://github.com/adieyal/sd-dynamic-prompts/issues/601 # noqa: B905,E501
        ]


def _is4xx(status: int):
    return status >= 400 and status < 500



import unittest

import flask
from sqlalchemy import create_engine
from sqlalchemy.dialects import registry

import radio_duck
from radio_duck import OperationalError
from radio_duck.sqlalchemy import RadioDuckDialect

http_server_port = 9000

registry.register(
    "radio_duck.district5", "radio_duck.sqlalchemy", "RadioDuckDialect"
)

url = (
    "radio_duck+district5://user:pass@localhost:"
    + str(http_server_port)  # noqa: W503
    + "/?api=/v1/sql/&scheme=http"  # noqa: W503
)

engine = create_engine(url)


def test_get_driver_name():
    dialect = RadioDuckDialect()
    assert dialect.name == "radio_duck"
    assert dialect.driver == "district5"


def test_get_isolation():
    dialect = RadioDuckDialect()
    assert dialect.isolation_level == "SNAPSHOT"
    assert dialect.get_isolation_level(None) == "SNAPSHOT"


def test_connect():
    from http_server_mock import HttpServerMock

    app = HttpServerMock(__name__)

    @app.route("/v1/sql/", methods=["POST"])
    def index():
        return (
            """ {
    "schema": [
        "NUMBER"
    ],
    "columns": [
        "count_star"
    ],
    "rows": [[5]] }""",
            200,
        )

    dialect = RadioDuckDialect(dbapi=radio_duck)
    with app.run("localhost", http_server_port):
        with dialect.connect(
            host="localhost",
            port=http_server_port,
            api="/v1/sql/",
            scheme="http",
        ) as conn:
            with conn.cursor() as cursor:
                cursor.execute("select count(*) as count_star from pond")
                row = cursor.fetchone()
                assert 5 == row[0]


def test_ping():
    from http_server_mock import HttpServerMock

    app = HttpServerMock(__name__)

    @app.route("/v1/sql/", methods=["POST"])
    def index():
        return "{}", 200

    with app.run("localhost", http_server_port):
        dialect = RadioDuckDialect(dbapi=radio_duck)
        with dialect.connect(
            host="localhost",
            port=http_server_port,
            api="/v1/sql/",
            scheme="http",
        ) as conn:
            dialect.do_ping(conn)


def test_is_disconnect():
    from http_server_mock import HttpServerMock

    app = HttpServerMock(__name__)

    @app.route("/v1/sql/", methods=["POST"])
    def index():
        return "{}", 200

    dialect = RadioDuckDialect(dbapi=radio_duck)
    assert not dialect.is_disconnect(None, None, None)
    assert not dialect.is_disconnect(OperationalError(msg="hi"), None, None)
    assert dialect.is_disconnect(
        OperationalError(msg=radio_duck.db.connect_close_resource_msg),
        None,
        None,
    )


def test_has_index():
    from http_server_mock import HttpServerMock

    app = HttpServerMock(__name__)

    @app.route("/v1/sql/", methods=["POST"])
    def index():
        json = flask.request.json
        request_idx_name = json["parameters"][2]
        response_rows = (
            '"rows":[["title_idx"]]'
            if request_idx_name == "title_idx"
            else '"rows":[]'
        )
        resp = f"""{{
          "schema": [
            "STRING"
          ],
          "columns": [
            "indexname"
          ],
          {response_rows}
        }}"""
        print(resp)
        resp = resp.replace("\n", " ")
        return resp

    with app.run("localhost", http_server_port):
        with engine.connect() as conn:
            dialect = RadioDuckDialect(dbapi=radio_duck)
            assert dialect.has_index(conn, "films", "title_idx", "main")
            assert not dialect.has_index(conn, "films", "unknown_idx", "main")


def test_has_table():
    from http_server_mock import HttpServerMock

    app = HttpServerMock(__name__)

    @app.route("/v1/sql/", methods=["POST"])
    def index():
        json = flask.request.json
        request_table_name = json["parameters"][1]
        response_rows = (
            '"rows":[["pond"]]'
            if request_table_name == "pond"
            else '"rows":[]'
        )
        resp = f"""{{
          "schema": [
            "string"
          ],
          "columns": [
            "table_name"
          ],
          {response_rows}
        }}"""
        print(resp)
        resp = resp.replace("\n", " ")
        return resp

    with app.run("localhost", http_server_port):
        with engine.connect() as conn:
            dialect = RadioDuckDialect(dbapi=radio_duck)
            assert dialect.has_table(conn, "pond", "main")
            assert not dialect.has_table(conn, "foobar", "main")


def test_has_seq():
    from http_server_mock import HttpServerMock

    app = HttpServerMock(__name__)

    @app.route("/v1/sql/", methods=["POST"])
    def index():
        json = flask.request.json
        request_table_name = json["parameters"][1]
        response_rows = (
            '"rows":[["serial"]]'
            if request_table_name == "serial"
            else '"rows":[]'
        )
        resp = f"""{{
          "schema": [
            "string"
          ],
          "columns": [
            "sequencename"
          ],
          {response_rows}
        }}"""
        print(resp)
        resp = resp.replace("\n", " ")
        return resp

    with app.run("localhost", http_server_port):
        with engine.connect() as conn:
            dialect = RadioDuckDialect(dbapi=radio_duck)
            assert dialect.has_sequence(conn, "serial", "main")
            assert not dialect.has_sequence(conn, "foobar", "main")


def test_get_tables():
    from http_server_mock import HttpServerMock

    app = HttpServerMock(__name__)

    @app.route("/v1/sql/", methods=["POST"])
    def index():
        json = flask.request.json
        request_schema_name = json["parameters"][0]
        response_rows = (
            '"rows":[["pond"]]'
            if request_schema_name == "main"
            else '"rows":[]'
        )
        resp = f"""{{
          "schema": [
            "string"
          ],
          "columns": [
            "tablename"
          ],
          {response_rows}
        }}"""
        print(resp)
        resp = resp.replace("\n", " ")
        return resp

    with app.run("localhost", http_server_port):
        with engine.connect() as conn:
            dialect = RadioDuckDialect(dbapi=radio_duck)
            assert ["pond"] == dialect.get_table_names(conn)


def test_get_views():
    from http_server_mock import HttpServerMock

    app = HttpServerMock(__name__)

    @app.route("/v1/sql/", methods=["POST"])
    def index():
        json = flask.request.json
        request_schema_name = json["parameters"][0]
        response_rows = (
            '"rows":[["pond_view"]]'
            if request_schema_name == "main"
            else '"rows":[]'
        )
        resp = f"""{{
          "schema": [
            "string"
          ],
          "columns": [
            "view_name"
          ],
          {response_rows}
        }}"""
        print(resp)
        resp = resp.replace("\n", " ")
        return resp

    with app.run("localhost", http_server_port):
        with engine.connect() as conn:
            dialect = RadioDuckDialect(dbapi=radio_duck)
            assert ["pond_view"] == dialect.get_view_names(conn)
            assert [] == dialect.get_view_names(conn, "foobar")


def test_get_view_sql():
    from http_server_mock import HttpServerMock

    app = HttpServerMock(__name__)

    @app.route("/v1/sql/", methods=["POST"])
    def index():
        json = flask.request.json
        request_view_name = json["parameters"][1]
        response_rows = (
            '"rows":[["CREATE VIEW pond_view AS SELECT * FROM pond"]]'
            if request_view_name == "pond_view"
            else '"rows":[]'
        )
        resp = f"""{{
          "schema": [
            "string"
          ],
          "columns": [
            "sql"
          ],
          {response_rows}
        }}"""
        print(resp)
        resp = resp.replace("\n", " ")
        return resp

    with app.run("localhost", http_server_port):
        with engine.connect() as conn:
            dialect = RadioDuckDialect(dbapi=radio_duck)
            expected = "CREATE VIEW pond_view AS SELECT * FROM pond"
            actual = dialect.get_view_definition(conn, "pond_view")
            assert expected == actual
            assert "" == dialect.get_view_definition(conn, "foobar")


def test_get_uniq_constraints():
    from http_server_mock import HttpServerMock

    app = HttpServerMock(__name__)

    @app.route("/v1/sql/", methods=["POST"])
    def index():
        json = flask.request.json
        table_name = json["parameters"][1]
        response_rows = (
            '"rows":[ ["UNIQUE(id, id)", ["id"] ] ]'
            if table_name == "students"
            else '"rows":[]'
        )
        resp = f"""{{
          "schema": [
            "string","list"
          ],
          "columns": [
            "name", "column_names"
          ],
          {response_rows}
        }}"""
        print(resp)
        resp = resp.replace("\n", " ")
        return resp

    with app.run("localhost", http_server_port):
        with engine.connect() as conn:
            dialect = RadioDuckDialect(dbapi=radio_duck)
            assert [
                {"name": "UNIQUE(id, id)", "column_names": ["id"]}
            ] == dialect.get_unique_constraints(conn, "students")
            assert [] == dialect.get_unique_constraints(conn, "foobar")


def test_get_temp_views():
    from http_server_mock import HttpServerMock

    app = HttpServerMock(__name__)

    @app.route("/v1/sql/", methods=["POST"])
    def index():
        response_rows = '"rows":[["temp_view"]]'
        resp = f"""{{
          "schema": [
            "string"
          ],
          "columns": [
            "view_name"
          ],
          {response_rows}
        }}"""
        print(resp)
        resp = resp.replace("\n", " ")
        return resp

    with app.run("localhost", http_server_port):
        with engine.connect() as conn:
            dialect = RadioDuckDialect(dbapi=radio_duck)
            assert ["temp_view"] == dialect.get_temp_view_names(conn, "main")
            assert ["temp_view"] == dialect.get_temp_view_names(conn, "foobar")
            assert ["temp_view"] == dialect.get_temp_view_names(conn)


def test_get_temp_tables():
    from http_server_mock import HttpServerMock

    app = HttpServerMock(__name__)

    @app.route("/v1/sql/", methods=["POST"])
    def index():
        json = flask.request.json
        schema_name = json["parameters"][0]
        response_rows = (
            '"rows":[ ["temp_table"] ]'
            if schema_name == "main"
            else '"rows":[]'
        )
        resp = f"""{{
          "schema": [
            "string"
          ],
          "columns": [
            "table_name"
          ],
          {response_rows}
        }}"""
        print(resp)
        resp = resp.replace("\n", " ")
        return resp

    with app.run("localhost", http_server_port):
        with engine.connect() as conn:
            dialect = RadioDuckDialect(dbapi=radio_duck)
            assert ["temp_table"] == dialect.get_temp_table_names(conn)
            assert [] == dialect.get_temp_table_names(conn, "foobar")


def test_get_seqs():
    from http_server_mock import HttpServerMock

    app = HttpServerMock(__name__)

    @app.route("/v1/sql/", methods=["POST"])
    def index():
        json = flask.request.json
        schema_name = json["parameters"][0]
        response_rows = (
            '"rows":[ ["seq1"] ]' if schema_name == "main" else '"rows":[]'
        )
        resp = f"""{{
          "schema": [
            "string"
          ],
          "columns": [
            "sequencename"
          ],
          {response_rows}
        }}"""
        print(resp)
        resp = resp.replace("\n", " ")
        return resp

    with app.run("localhost", http_server_port):
        with engine.connect() as conn:
            dialect = RadioDuckDialect(dbapi=radio_duck)
            assert ["seq1"] == dialect.get_sequence_names(conn)
            assert [] == dialect.get_sequence_names(conn, "foobar")


def test_get_pk():
    from http_server_mock import HttpServerMock

    app = HttpServerMock(__name__)

    @app.route("/v1/sql/", methods=["POST"])
    def index():
        json = flask.request.json
        table_name = json["parameters"][1]
        response_rows = (
            '"rows":[ ["PRIMARY KEY(id)", ["id"] ] ]'
            if table_name == "students"
            else '"rows":[]'
        )
        resp = f"""{{
          "schema": [
            "string","list"
          ],
          "columns": [
            "name", "column_names"
          ],
          {response_rows}
        }}"""
        print(resp)
        resp = resp.replace("\n", " ")
        return resp

    with app.run("localhost", http_server_port):
        with engine.connect() as conn:
            dialect = RadioDuckDialect(dbapi=radio_duck)
            assert [
                {"name": "PRIMARY KEY(id)", "column_names": ["id"]}
            ] == dialect.get_pk_constraint(conn, "students")
            assert [] == dialect.get_pk_constraint(conn, "foobar")


def test_get_indexes():
    from http_server_mock import HttpServerMock

    app = HttpServerMock(__name__)

    @app.route("/v1/sql/", methods=["POST"])
    def index():
        json = flask.request.json
        table_name = json["parameters"][1]
        response_rows = (
            '"rows":[ ["city_id_ndex", "CREATE INDEX city_id_ndex ON employee (city,id);", false] ]'  # noqa: E501,B950
            if table_name == "employee"
            else '"rows":[]'
        )
        resp = f"""{{
          "schema": [
            "string","string","bool"
          ],
          "columns": [
            "index_name","sql","is_unique"
          ],
          {response_rows}
        }}"""
        print(resp)
        resp = resp.replace("\n", " ")
        return resp

    with app.run("localhost", http_server_port):
        with engine.connect() as conn:
            dialect = RadioDuckDialect(dbapi=radio_duck)
            assert [
                {
                    "name": "city_id_ndex",
                    "column_names": ["city", "id"],
                    "unique": False,
                }
            ] == dialect.get_indexes(conn, "employee")
            assert [] == dialect.get_indexes(conn, "foobar")


def test_get_fk():
    from http_server_mock import HttpServerMock

    app = HttpServerMock(__name__)

    @app.route("/v1/sql/", methods=["POST"])
    def index():
        json = flask.request.json
        table_name = json["parameters"][1]
        response_rows = (
            '"rows":[ ["FOREIGN KEY (t1_id) REFERENCES t1(id)", ["t1_id"] ] ]'
            if table_name == "t2"
            else '"rows":[]'
        )
        resp = f"""{{
          "schema": [
            "string","list"
          ],
          "columns": [
            "name", "column_names"
          ],
          {response_rows}
        }}"""
        print(resp)
        resp = resp.replace("\n", " ")
        return resp

    with app.run("localhost", http_server_port):
        with engine.connect() as conn:
            dialect = RadioDuckDialect(dbapi=radio_duck)
            assert [
                {
                    "name": "FOREIGN KEY (t1_id) REFERENCES t1(id)",
                    "column_names": ["t1_id"],
                }
            ] == dialect.get_foreign_keys(conn, "t2")
            assert [] == dialect.get_pk_constraint(conn, "foobar")


def test_get_columns():
    from http_server_mock import HttpServerMock

    app = HttpServerMock(__name__)

    @app.route("/v1/sql/", methods=["POST"])
    def index():
        response_rows = (
            '"rows":[ [1, "map_col", "MAP(INTEGER, DOUBLE)", false, null,'
            ' false], [2, "listing", "INTEGER[]", false, null, false], [3,'
            ' "unionn", "UNION(num INTEGER, str VARCHAR)", false, null,'
            ' false], [4, "name", "STRING", true, "acb", true]]'
        )
        resp = f"""{{
          "schema": [
                "NUMBER",
                "STRING",
                "STRING",
                "bool",
                "STRING",
                "bool"
          ],
          "columns": [
            "cid",
            "name",
            "type",
            "notnull",
            "dflt_value",
            "pk"
          ],
          {response_rows}
        }}"""
        print(resp)
        resp = resp.replace("\n", " ")
        return resp

    with app.run("localhost", http_server_port):
        with engine.connect() as conn:
            from sqlalchemy import types

            dialect = RadioDuckDialect(dbapi=radio_duck)
            result = dialect.get_columns(conn, "pond", "main")
            assert [
                {
                    "name": "map_col",
                    "type": types.BLOB,
                    "nullable": True,
                    "default": None,
                    "primary_key": False,
                },
                {
                    "name": "listing",
                    "type": types.ARRAY,
                    "nullable": True,
                    "default": None,
                    "primary_key": False,
                },
                {
                    "name": "unionn",
                    "type": types.BLOB,
                    "nullable": True,
                    "default": None,
                    "primary_key": False,
                },
                {
                    "name": "name",
                    "type": types.String,
                    "nullable": False,
                    "default": "acb",
                    "primary_key": True,
                },
            ] == result


def test_get_checks():
    from http_server_mock import HttpServerMock

    app = HttpServerMock(__name__)

    @app.route("/v1/sql/", methods=["POST"])
    def index():
        json = flask.request.json
        table_name = json["parameters"][1]
        response_rows = (
            '"rows":[ ["CHECK((x<y))" , "(x<y)" ] ]'
            if table_name == "t2"
            else '"rows":[]'
        )
        resp = f"""{{
          "schema": [
            "string","string"
          ],
          "columns": [
            "name", "sqltext"
          ],
          {response_rows}
        }}"""
        print(resp)
        resp = resp.replace("\n", " ")
        return resp

    with app.run("localhost", http_server_port):
        with engine.connect() as conn:
            dialect = RadioDuckDialect(dbapi=radio_duck)
            assert [
                {"name": "CHECK((x<y))", "sqltext": "(x<y)"}
            ] == dialect.get_check_constraints(conn, "t2")
            assert [] == dialect.get_check_constraints(conn, "foobar")


def test_get_schemas():
    from http_server_mock import HttpServerMock

    app = HttpServerMock(__name__)

    @app.route("/v1/sql/", methods=["POST"])
    def index():
        response_rows = '"rows":[["main"]]'
        resp = f"""{{
          "schema": [
            "string"
          ],
          "columns": [
            "schema_name"
          ],
          {response_rows}
        }}"""
        print(resp)
        resp = resp.replace("\n", " ")
        return resp

    with app.run("localhost", http_server_port):
        with engine.connect() as conn:
            dialect = RadioDuckDialect(dbapi=radio_duck)
            assert ["main"] == dialect.get_schema_names(conn)


class NotImplementedTest(unittest.TestCase):
    def test_get_table_comment(self):
        dialect = RadioDuckDialect(dbapi=radio_duck)
        with self.assertRaises(NotImplementedError):
            dialect.get_table_comment(None, None)



# https://peps.python.org/pep-0249/#globals

from typing import Any

from radio_duck.db import Connection

apilevel = "2.0"

threadsafety = 1

paramstyle = "qmark"

__all__ = ["apilevel", "threadsafety", "paramstyle", "connect"]


# ----------------------------------------------------------
# Functions


def connect(*args, **kwargs) -> Any:
    """
    Connect to the database
    :param args:
    :param kwargs: minimum kwargs are host,port,
    api(endpoint url ex: '/v1/sql'). optional: timeout_sec
    :return: Connection object
    :raise ProgrammingError on incorrect scheme
    :raise OperationalError if unable to connect to database
    """
    return Connection(*args, **kwargs)



from __future__ import annotations

import logging
import re
from typing import TYPE_CHECKING

import sqlalchemy

from radio_duck.reserved_keywords import keyword_list

if TYPE_CHECKING:
    from _typeshed import DBAPIConnection

from typing import Any, Callable, List

# https://docs.sqlalchemy.org/en/20/core/internals.html#sqlalchemy.engine.Dialect.do_terminate
# from sqlalchemy.engine import default, interfaces
from sqlalchemy.engine import default
from sqlalchemy.sql import compiler

import radio_duck
from radio_duck import db_types
from radio_duck.db import connect_close_resource_msg
from radio_duck.exceptions import NotSupportedError
from radio_duck.queries import (
    get_columns,
    get_constraints,
    get_indexes,
    get_schemas,
    get_sequences,
    get_tables,
    get_temp_tables,
    get_temp_views,
    get_view_sql,
    get_views,
    has_index_query,
    has_sequence_query,
    has_table_query,
)


class RadioDuckDialectPreparer(compiler.IdentifierPreparer):
    reserved_words = keyword_list


class RadioDuckDialectTypeCompiler(compiler.GenericTypeCompiler):
    """
    Refer
    https://duckdb.org/docs/sql/data_types/numeric
    https://duckdb.org/docs/sql/data_types/overview.html
    https://www.postgresql.org/docs/current/datatype-numeric.html
    """

    visit_REAL = compiler.GenericTypeCompiler.visit_FLOAT
    visit_NUMERIC = compiler.GenericTypeCompiler.visit_DECIMAL

    visit_DATETIME = compiler.GenericTypeCompiler.visit_TIMESTAMP

    visit_CLOB = compiler.GenericTypeCompiler.visit_BLOB
    visit_NCLOB = compiler.GenericTypeCompiler.visit_BLOB
    visit_BINARY = compiler.GenericTypeCompiler.visit_BLOB
    visit_VARBINARY = compiler.GenericTypeCompiler.visit_BLOB

    visit_TEXT = compiler.GenericTypeCompiler.visit_VARCHAR


class RadioDuckDialect(default.DefaultDialect):
    type_compiler = RadioDuckDialectTypeCompiler
    preparer = RadioDuckDialectPreparer

    supports_sequences = True
    supports_native_enum = True
    supports_native_boolean = True

    tuple_in_values = True

    # if the NUMERIC type
    # returns decimal.Decimal.
    # *not* the FLOAT type however.
    supports_native_decimal = True

    name = "radio_duck"
    driver = "district5"

    isolation_level = "SNAPSHOT"

    default_paramstyle = "qmark"

    # not sure if this is a real thing but the compiler will deliver it
    # if this is the only flag enabled.
    supports_empty_insert = False
    """dialect supports INSERT () VALUES ()"""

    supports_multivalues_insert = True

    default_schema_name = "main"  # duckdb default schema is main

    @classmethod
    def dbapi(cls):
        return radio_duck

    def __init__(
        self,
        convert_unicode=False,
        encoding="utf-8",
        paramstyle=None,
        dbapi=None,
        implicit_returning=None,
        case_sensitive=True,
        supports_native_boolean=None,
        max_identifier_length=None,
        label_length=None,
        # int() is because the @deprecated_params decorator cannot accommodate
        # the direct reference to the "NO_LINTING" object
        compiler_linting=int(compiler.NO_LINTING),  # noqa: B008
        server_side_cursors=False,
        **kwargs,
    ):
        super().__init__(
            convert_unicode=convert_unicode,
            encoding=encoding,
            paramstyle=paramstyle,
            dbapi=dbapi,
            implicit_returning=implicit_returning,
            case_sensitive=case_sensitive,
            supports_native_boolean=supports_native_boolean,
            max_identifier_length=max_identifier_length,
            label_length=label_length,
            compiler_linting=compiler_linting,
            server_side_cursors=server_side_cursors,
            **kwargs,
        )

    # do methods

    def do_savepoint(self, connection, name):
        raise NotImplementedError()

    def do_rollback_to_savepoint(self, connection, name):
        raise NotImplementedError()

    def do_rollback(self, dbapi_connection):
        pass
        # raise NotImplementedError()

    def do_release_savepoint(self, connection, name):
        raise NotImplementedError()

    def do_recover_twophase(self, connection):
        raise NotImplementedError()

    def do_prepare_twophase(self, connection, xid):
        raise NotImplementedError()

    def do_commit_twophase(
        self, connection, xid, is_prepared=True, recover=False
    ):
        raise NotImplementedError()

    def do_commit(self, dbapi_connection):
        pass

    def do_begin_twophase(self, connection, xid) -> None:
        raise NotImplementedError()

    def do_begin(self, dbapi_connection):
        pass

    # -- connect methods

    def create_connect_args(self, url):
        """Build DB-API compatible connection arguments.
        :param url: a :class:`.URL` object

        :return: a tuple of ``(*args, **kwargs)`` which will be passed to the
         :meth:`.Dialect.connect` method.

        .. seealso::    :meth:`.URL.translate_connect_args`
        """

        opts = url.translate_connect_args()
        opts.update(url.query)
        # todo: default impl returns [[], opts]..
        # this is a bug as pydoc says return tuple! not list
        return [], opts

    def create_xid(self):
        raise NotSupportedError("transactions not supported over http yet")

    @classmethod
    def engine_created(cls, engine):
        super().engine_created(engine)
        logging.info("radio_duck dialect engine created")

    def reset_isolation_level(self, dbapi_conn) -> None:
        pass

    def set_isolation_level(self, dbapi_conn, level) -> None:
        pass

    def on_connect(self) -> Callable[[DBAPIConnection], object] | None:
        def do_on_connect(connection):
            # todo. set duckdb specific flags
            # connection.execute("SET SPECIAL FLAGS etc")
            logging.debug("radio_duck pre connection  establishment hook.")

        return do_on_connect

    def on_connect_url(
        self, url
    ) -> Callable[[DBAPIConnection], object] | None:
        def do_on_connect_url(connection):
            # todo. set duckdb specific flags
            # connection.execute("SET SPECIAL FLAGS etc")
            logging.debug(
                "radio_duck pre connection establishment hook to url: ", url
            )

        return do_on_connect_url

    def is_disconnect(self, e, connection, cursor):
        """
            Return True if the given DB-API error
            indicates an invalid connection
        :param e:
        :param connection:
        :param cursor:
        :return:
        """
        # return all(
        #     e is not None,
        #     radio_duck.db.connect_close_resource_msg in str(e)
        # )
        return False if e is None else connect_close_resource_msg in str(e)

    # ----has methods

    def has_index(self, connection, table_name, index_name, schema=None):
        if schema is None or "" == schema.strip():
            schema = RadioDuckDialect.default_schema_name
        rows = self._execute_query(
            connection, has_index_query, schema, table_name, index_name
        )
        return len(rows) == 1

    def has_table(self, connection, table_name, schema=None, **kw) -> None:
        if schema is None or "" == schema.strip():
            schema = RadioDuckDialect.default_schema_name
        rows = self._execute_query(
            connection, has_table_query, schema, table_name
        )
        return len(rows) == 1

    def has_sequence(
        self, connection, sequence_name, schema=None, **kw
    ) -> bool:
        if schema is None or "" == schema.strip():
            schema = RadioDuckDialect.supports_schemas
        rows = self._execute_query(
            connection, has_sequence_query, schema, sequence_name
        )
        return len(rows) == 1

    # ----get methods

    def get_table_names(
        self, connection, schema=None, **kw
    ) -> List[str] | None:
        if schema is None or "" == schema.strip():
            schema = RadioDuckDialect.default_schema_name

        rows = self._execute_query(
            connection, get_tables, schema
        )  # list of list
        table_names = [col_val for row in rows for col_val in row]
        return table_names

    def get_view_names(self, connection, schema=None, **kw):
        if schema is None or "" == schema.strip():
            schema = RadioDuckDialect.default_schema_name
        rows = self._execute_query(
            connection, get_views, schema
        )  # list of list
        views = [col_val for row in rows for col_val in row]
        return views

    def get_view_definition(
        self, connection, view_name, schema=None, **kw
    ) -> None:
        if schema is None or "" == schema.strip():
            schema = RadioDuckDialect.default_schema_name
        rows = self._execute_query(connection, get_view_sql, schema, view_name)
        if len(rows) == 0:
            return ""
        row = rows[0]
        return row[0]  # return sql

    def get_unique_constraints(
        self, connection, table_name, schema=None, **kw
    ) -> list[dict[str, Any]]:
        if schema is None or "" == schema.strip():
            schema = RadioDuckDialect.default_schema_name

        rows = self._execute_query(
            connection, get_constraints, schema, table_name, "UNIQUE"
        )  # list of list
        list_of_maps = [
            {"name": row[0], "column_names": row[1]} for row in rows
        ]
        return list_of_maps

    def get_temp_view_names(self, connection, schema=None, **kw):
        # Temporary views exist in a special schema, so a schema name
        # cannot be given when creating a temporary view.
        # The name of the view must be distinct from the name of any
        # other view or table in the same schema.
        # hence schema is ignored for query
        rows = self._execute_query(connection, get_temp_views)  # list of list
        temp_views = [col_val for row in rows for col_val in row]
        return temp_views

    def get_temp_table_names(self, connection, schema=None, **kw):
        if schema is None or "" == schema.strip():
            schema = RadioDuckDialect.default_schema_name

        rows = self._execute_query(
            connection, get_temp_tables, schema
        )  # list of list
        temp_tables = [col_val for row in rows for col_val in row]
        return temp_tables

    def get_table_comment(self, connection, table_name, schema=None, **kw):
        raise NotImplementedError()

    def get_sequence_names(self, connection, schema=None, **kw) -> list[Any]:
        if schema is None or "" == schema.strip():
            schema = RadioDuckDialect.default_schema_name
        rows = self._execute_query(
            connection, get_sequences, schema
        )  # list of list
        seq_names = [col_val for row in rows for col_val in row]
        return seq_names

    def get_pk_constraint(self, connection, table_name, schema=None, **kw):
        if schema is None or "" == schema.strip():
            schema = RadioDuckDialect.default_schema_name

        rows = self._execute_query(
            connection, get_constraints, schema, table_name, "PRIMARY KEY"
        )  # list of list
        if len(rows) == 0:
            return []
        row = rows[0]
        list_of_maps = (
            [{"name": row[0], "column_names": row[1]}]
            if row is not None
            else []
        )
        return list_of_maps

    def get_indexes(self, connection, table_name, schema=None, **kw):
        if schema is None or "" == schema.strip():
            schema = RadioDuckDialect.default_schema_name

        rows = self._execute_query(
            connection, get_indexes, schema, table_name
        )  # list of list
        list_of_maps = [
            {"name": row[0], "sql": row[1], "unique": row[2]} for row in rows
        ]
        for mp in list_of_maps:
            # add column names
            mp["column_names"] = []
            # parse sql for column names
            sql = mp["sql"]
            match = re.search(r"\((.*?)\)", sql)
            if match:
                columns = match.group(1).split(",")
                columns = [column.strip() for column in columns]
                mp["column_names"] = columns
            del mp["sql"]

        return list_of_maps

    def get_foreign_keys(self, connection, table_name, schema=None, **kw):
        if schema is None or "" == schema.strip():
            schema = RadioDuckDialect.default_schema_name

        rows = self._execute_query(
            connection, get_constraints, schema, table_name, "FOREIGN KEY"
        )  # list of list
        list_of_maps = [
            {"name": row[0], "column_names": row[1]} for row in rows
        ]
        return list_of_maps

    def get_columns(self, connection, table_name, schema=None, **kw):
        if schema is None or "" == schema.strip():
            schema = RadioDuckDialect.default_schema_name

        # query = f"SET schema '{schema}'; PRAGMA table_info('{table_name}')"
        query = get_columns.format(schema, table_name)
        rows = self._execute_query(connection, query)  # list of list
        list_of_maps = [
            {
                "name": row[1],
                "type": db_types.get_alchemy_type(row[2]),
                "nullable": not row[3],  # outputcolumn is 'notnull'
                "default": row[4],
                # "autoincrement": "",
                "primary_key": row[5],
            }
            for row in rows
        ]
        # not sending 'autoincrement' & 'sequence'
        # duckdb does not have these, hence not in result
        return list_of_maps

    def get_check_constraints(
        self, connection, table_name, schema=None, **kw
    ) -> list[dict[str, Any]]:
        if schema is None or "" == schema.strip():
            schema = RadioDuckDialect.default_schema_name

        rows = self._execute_query(
            connection, get_constraints, schema, table_name, "CHECK"
        )  # list of list
        list_of_maps = [{"name": row[0], "sqltext": row[1]} for row in rows]
        # not including keys 'autoincrement' & 'sequence' -
        # duckdb does not have these, hence not in result
        return list_of_maps

    def get_isolation_level(self, dbapi_conn) -> str | None:
        return self.isolation_level

    def get_default_isolation_level(self, dbapi_conn):
        # we have only snapshot
        return self.isolation_level

    def get_schema_names(
        self, connection: sqlalchemy.engine.base.Connection, **kw
    ):
        """
        superset ui asks the dialect to list schemas too
        this is not part of alchemy specification
        :param connection: alchemy base connection
        :param kw:
        :return list of schema names where
        catalog_name is neither temp not system:
        """
        rows = self._execute_query(connection, get_schemas)
        return [val for row in rows for val in row]

    def _execute_query(
        self,
        connection: sqlalchemy.engine.base.Connection,
        query: str,
        *params,
        **kw,
    ):
        if not params:
            result = connection.execute(query)
        else:
            result = connection.execute(
                query, params
            )  # use qmark ? style for params
        return result.fetchall()  # list of list



# taken from
# https://github.com/duckdb/duckdb/blob/ff7f24fd8e3128d94371827523dae85ebaf58713/third_party/libpg_query/grammar/keywords/reserved_keywords.list
keyword_list = [
    "ALL",
    "ANALYSE",
    "ANALYZE",
    "AND",
    "ANY",
    "ARRAY",
    "AS",
    "ASC_P",
    "ASYMMETRIC",
    "BOTH",
    "CASE",
    "CAST",
    "CHECK_P",
    "COLLATE",
    "COLUMN",
    "CONSTRAINT",
    "CREATE_P",
    "DEFAULT",
    "DEFERRABLE",
    "DESC_P",
    "DISTINCT",
    "DO",
    "ELSE",
    "END_P",
    "EXCEPT",
    "FALSE_P",
    "FETCH",
    "FOR",
    "FOREIGN",
    "FROM",
    "GRANT",
    "GROUP_P",
    "HAVING",
    "QUALIFY",
    "IN_P",
    "INITIALLY",
    "INTERSECT",
    "INTO",
    "LATERAL_P",
    "LEADING",
    "LIMIT",
    "NOT",
    "NULL_P",
    "OFFSET",
    "ON",
    "ONLY",
    "OR",
    "ORDER",
    "PIVOT",
    "PIVOT_WIDER",
    "PIVOT_LONGER",
    "PLACING",
    "PRIMARY",
    "REFERENCES",
    "RETURNING",
    "SELECT",
    "SOME",
    "SYMMETRIC",
    "TABLE",
    "THEN",
    "TO",
    "TRAILING",
    "TRUE_P",
    "UNION",
    "UNIQUE",
    "UNPIVOT",
    "USING",
    "VARIADIC",
    "WHEN",
    "WHERE",
    "WINDOW",
    "WITH",
]



import datetime
import time

from sqlalchemy import types as alchemy_types

# ----------------------------------------------------------
# Types

#: The type for date values.
Date = datetime.date
#: The type for time values.
Time = datetime.time
#: The type for timestamp values.
Timestamp = datetime.datetime


def Binary(string):
    return str(string)


def DateFromTicks(ticks: int) -> Date:
    """Construct a date value from a count of seconds."""
    # Standard implementations from PEP 249 itself
    return Date(*time.localtime(ticks)[:3])


def TimeFromTicks(ticks: int) -> Time:
    """Construct a time value from a count of seconds."""
    return Time(*time.localtime(ticks)[3:6])


def TimestampFromTicks(ticks: int) -> Timestamp:
    """Construct a timestamp value from a count of seconds."""
    return Timestamp(*time.localtime(ticks)[:6])


class Type(object):
    def __init__(self, type_code: int):
        self._type_code = type_code

    def get_type_code(self):
        return self._type_code


# use the 5 ducks
# from https://disney.fandom.com/wiki/District_5_Ducks
# using jersey numbers :)
STRING = Type(96)  # one of 5 ducks in flying v -charlie convay.
BINARY = Type(4)  # one of 5 ducks in flying v -averman
NUMBER = Type(9)  # one of 5 ducks in flying v -jesse hall.
DATETIME = Type(1)  # one of 5 ducks in flying v -Terry Hall
ROWID = Type(0)  # one of 5 ducks in flying v -Guy Germaine
UNKNOWN = Type(-1)  # one of 5 ducks in flying v -Guy Germaine

__types = {
    "STRING": STRING,
    "BINARY": BINARY,
    "NUMBER": NUMBER,
    "DATETIME": DATETIME,
    "ROWID": ROWID,
}


def get_type_code(col_type: str) -> int:
    return __types.get(col_type, UNKNOWN).get_type_code()


# ----------------------------------------------------------


# https://github.com/preset-io/elasticsearch-dbapi/blob/master/es/baseapi.py

_alchemy_type_map = {
    "BIGINT": alchemy_types.BIGINT,
    "INT8": alchemy_types.BIGINT,
    "LONG": alchemy_types.BIGINT,
    "BIT": alchemy_types.String,
    "BITSTRING": alchemy_types.String,
    "BOOLEAN": alchemy_types.Boolean,
    "BOOL": alchemy_types.Boolean,
    "LOGICAL": alchemy_types.Boolean,
    "BLOB": alchemy_types.BLOB,
    "BYTEA": alchemy_types.BLOB,
    "BINARY": alchemy_types.BLOB,
    "VARBINARY": alchemy_types.BLOB,
    "DATE": alchemy_types.DATE,
    "DOUBLE": alchemy_types.FLOAT,
    "FLOAT8": alchemy_types.FLOAT,
    "NUMERIC": alchemy_types.FLOAT,
    "DECIMAL": alchemy_types.FLOAT,
    "DECIMAL(PREC, SCALE)": alchemy_types.DECIMAL,
    "HUGEINT": alchemy_types.BIGINT,
    "INTEGER": alchemy_types.Integer,
    "INT": alchemy_types.Integer,
    "INT4": alchemy_types.Integer,
    "SIGNED": alchemy_types.Integer,
    "INTERVAL": alchemy_types.Interval,
    "REAL": alchemy_types.FLOAT,
    "FLOAT4": alchemy_types.FLOAT,
    "FLOAT": alchemy_types.FLOAT,
    "SMALLINT": alchemy_types.SMALLINT,
    "INT2": alchemy_types.SMALLINT,
    "SHORT": alchemy_types.SMALLINT,
    "TIME": alchemy_types.TIME,
    "TIMESTAMP": alchemy_types.TIMESTAMP,
    "DATETIME": alchemy_types.TIMESTAMP,
    "TIMESTAMP WITH TIME ZONE": alchemy_types.TIMESTAMP,
    "TIMESTAMPTZ": alchemy_types.TIMESTAMP,
    "TINYINT": alchemy_types.SMALLINT,
    "INT1": alchemy_types.SMALLINT,
    "UBIGINT": alchemy_types.BIGINT,
    "UINTEGER": alchemy_types.INTEGER,
    "USMALLINT": alchemy_types.SMALLINT,
    "UTINYINT": alchemy_types.SMALLINT,
    "UUID": alchemy_types.String,
    "STRING": alchemy_types.String,
    "TEXT": alchemy_types.String,
    "BPCHAR": alchemy_types.String,
    "CHAR": alchemy_types.String,
    "VARCHAR": alchemy_types.String,
}


def get_alchemy_type(data_type):
    if "[]" in data_type:  # LIST
        return alchemy_types.ARRAY
    elif "UNION" in data_type or "STRUCT" in data_type or "MAP" in data_type:
        return alchemy_types.BLOB  # other complex
    else:
        return _alchemy_type_map[data_type.upper()]



from sqlalchemy import types as alchemy_types

from radio_duck import db_types


def test_duckdb_to_alchemy_types():
    test_cases = {
        "BIGINT": alchemy_types.BIGINT,
        "INT8": alchemy_types.BIGINT,
        "LONG": alchemy_types.BIGINT,
        "BIT": alchemy_types.String,
        "BITSTRING": alchemy_types.String,
        "BOOLEAN": alchemy_types.Boolean,
        "BOOL": alchemy_types.Boolean,
        "LOGICAL": alchemy_types.Boolean,
        "BLOB": alchemy_types.BLOB,
        "BYTEA": alchemy_types.BLOB,
        "BINARY": alchemy_types.BLOB,
        "VARBINARY": alchemy_types.BLOB,
        "DATE": alchemy_types.DATE,
        "DOUBLE": alchemy_types.FLOAT,
        "FLOAT8": alchemy_types.FLOAT,
        "DECIMAL": alchemy_types.FLOAT,
        "NUMERIC": alchemy_types.FLOAT,
        "DECIMAL(PREC, SCALE)": alchemy_types.DECIMAL,
        "HUGEINT": alchemy_types.BIGINT,
        "INTEGER": alchemy_types.Integer,
        "INT4": alchemy_types.Integer,
        "INT": alchemy_types.Integer,
        "SIGNED": alchemy_types.Integer,
        "INTERVAL": alchemy_types.Interval,
        "REAL": alchemy_types.FLOAT,
        "FLOAT4": alchemy_types.FLOAT,
        "FLOAT": alchemy_types.FLOAT,
        "SMALLINT": alchemy_types.SMALLINT,
        "INT2": alchemy_types.SMALLINT,
        "SHORT": alchemy_types.SMALLINT,
        "TIME": alchemy_types.TIME,
        "TIMESTAMP": alchemy_types.TIMESTAMP,
        "DATETIME": alchemy_types.TIMESTAMP,
        "TIMESTAMP WITH TIME ZONE": alchemy_types.TIMESTAMP,
        "TIMESTAMPTZ": alchemy_types.TIMESTAMP,
        "TINYINT": alchemy_types.SMALLINT,
        "INT1": alchemy_types.SMALLINT,
        "UINTEGER": alchemy_types.INTEGER,
        "USMALLINT": alchemy_types.SMALLINT,
        "UTINYINT": alchemy_types.SMALLINT,
        "UUID": alchemy_types.String,
        "VARCHAR": alchemy_types.String,
        "CHAR": alchemy_types.String,
        "BPCHAR": alchemy_types.String,
        "TEXT": alchemy_types.String,
        "STRING": alchemy_types.String,
        "UBIGINT": alchemy_types.BIGINT,
    }
    for duckdb_type in test_cases:
        expected = test_cases[duckdb_type]
        assert expected == db_types.get_alchemy_type(duckdb_type)
        assert len(test_cases) == len(db_types._alchemy_type_map)
